@@ -1,4 +1,3 @@
-#include <sys/ioctl.h>
 #include <sys/types.h>
 #include <sys/wait.h>
 
@@ -6,7 +5,6 @@
 #include <curses.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <limits.h>
 #include <locale.h>
 #include <signal.h>
 #include <stdarg.h>
@@ -19,15 +17,23 @@
 #include <unistd.h>
 #include <wchar.h>
 
-#define LEN(a) sizeof((a))/sizeof((a)[0])
+#define LEN(a)              sizeof((a))/sizeof((a)[0])
+#define PAD_TRUNCATE_SYMBOL "\xe2\x80\xa6" /* symbol: "ellipsis" */
+#define SCROLLBAR_SYMBOL    "\xe2\x94\x82" /* symbol: "light vertical" */
+
+enum {
+	ATTR_RESET = 0,
+	ATTR_BOLD_ON = 1, ATTR_FAINT_ON = 2,
+	ATTR_REVERSE_ON = 7, ATTR_REVERSE_OFF = 27,
+};
+
+enum Pane { PaneFeeds, PaneItems, PaneLast };
 
 enum {
 	FieldUnixTimestamp = 0, FieldTitle, FieldLink, FieldContent,
 	FieldContentType, FieldId, FieldAuthor, FieldEnclosure, FieldLast
 };
 
-enum Pane { PaneFeeds, PaneItems, PaneLast };
-
 struct win {
 	int width;
 	int height;
@@ -40,17 +46,6 @@ struct row {
 	void *data;
 };
 
-struct scrollbar {
-	int tickpos;
-	int ticksize;
-	int x; /* absolute x position of the window on the screen */
-	int y; /* absolute y position of the window on the screen */
-	int size; /* absolute size of the bar */
-	int focused; /* has focus or not */
-	int hidden; /* is visible or not */
-	int dirty; /* needs draw update */
-};
-
 struct pane {
 	int x; /* absolute x position of the window on the screen */
 	int y; /* absolute y position of the window on the screen */
@@ -68,6 +63,17 @@ struct pane {
 	int (*row_match)(struct pane *, struct row *, const char *);
 };
 
+struct scrollbar {
+	int tickpos;
+	int ticksize;
+	int x; /* absolute x position of the window on the screen */
+	int y; /* absolute y position of the window on the screen */
+	int size; /* absolute size of the bar */
+	int focused; /* has focus or not */
+	int hidden; /* is visible or not */
+	int dirty; /* needs draw update */
+};
+
 struct statusbar {
 	int x; /* absolute x position of the window on the screen */
 	int y; /* absolute y position of the window on the screen */
@@ -77,33 +83,7 @@ struct statusbar {
 	int dirty; /* needs draw update */
 };
 
-#undef err
-void err(int, const char *, ...);
-
-void alldirty(void);
-void cleanup(void);
-void draw(void);
-void pane_draw(struct pane *);
-void pane_setpos(struct pane *p, off_t pos);
-void statusbar_draw(struct statusbar *);
-void sighandler(int);
-void updategeom(void);
-void updatesidebar(int);
-
-static struct statusbar statusbar;
-static struct pane panes[PaneLast];
-static struct scrollbar scrollbars[PaneLast]; /* each pane has a scrollbar */
-static struct win win;
-static size_t selpane;
-
-static struct termios tsave; /* terminal state at startup */
-static struct termios tcur;
-
-static struct winsize winsz; /* window size information */
-static int ttyfd = 0; /* fd of tty */
-static int devnullfd;
-
-static int needcleanup;
+/* /UI */
 
 /* feed info */
 struct feed {
@@ -114,35 +94,55 @@ struct feed {
 	FILE *fp;               /* file pointer */
 };
 
-static struct feed *feeds;
-static struct feed *curfeed;
-static size_t nfeeds; /* amount of feeds */
-
-static time_t comparetime;
-
 struct item {
 	char *fields[FieldLast];
 	char *line; /* allocated split line */
 	time_t timestamp;
-	struct tm tm;
+	int timeok;
 	int isnew;
 	off_t offset; /* line offset in file */
 };
 
+#undef err
+void err(int, const char *, ...);
+
+void alldirty(void);
+void cleanup(void);
+void draw(void);
+void pane_draw(struct pane *);
+void sighandler(int);
+void updategeom(void);
+void updatesidebar(int);
+
+static struct statusbar statusbar;
+static struct pane panes[PaneLast];
+static struct scrollbar scrollbars[PaneLast]; /* each pane has a scrollbar */
+static struct win win;
+static size_t selpane;
+static int usemouse = 1; /* use xterm mouse tracking */
+static int onlynew = 0; /* show only new in sidebar */
+
+static struct termios tsave; /* terminal state at startup */
+static struct termios tcur;
+static int devnullfd;
+static int needcleanup;
+
+static struct feed *feeds;
+static struct feed *curfeed;
+static size_t nfeeds; /* amount of feeds */
+static time_t comparetime;
+
 /* config */
 
-/* use xterm mouse tracking */
-static int usemouse = 1;
-/* show only new in sidebar (default = 0 = all) */
-static int onlynew = 0;
 /* Allow to lazyload items when a file is specified? This saves memory but
    increases some latency when seeking items. It also causes issues if the
    feed is changed while having the UI open (and offsets are changed). */
 static int lazyload = 0;
 
-static char *plumber = "xdg-open";
-static char *piper = "less";
+static char *plumber = "xdg-open"; /* environment variable: $SFEED_PLUMBER */
+static char *piper = "less"; /* environment variable: $SFEED_PIPER */
 
+/* like BSD err(), but calls cleanup() and _exit(). */
 void
 err(int code, const char *fmt, ...)
 {
@@ -293,8 +293,8 @@ utf8pad(char *buf, size_t bufsiz, const char *s, size_t len, int pad)
 		if (col + w > len || (col + w == len && s[i + rl])) {
 			if (siz + 4 >= bufsiz)
 				return -1;
-			memcpy(&buf[siz], "\xe2\x80\xa6", 3); /* ellipsis */
-			siz += 3;
+			memcpy(&buf[siz], PAD_TRUNCATE_SYMBOL, sizeof(PAD_TRUNCATE_SYMBOL) - 1);
+			siz += sizeof(PAD_TRUNCATE_SYMBOL) - 1;
 			if (col + w == len && w > 1)
 				buf[siz++] = pad;
 			buf[siz] = '\0';
@@ -329,7 +329,7 @@ printpad(const char *s, int width)
 void
 resettitle(void)
 {
-	printf("\x1b%c", 'c'); /* reset title and state */
+	fputs("\x1b""c", stdout); /* rs1: reset title and state */
 }
 
 void
@@ -345,23 +345,92 @@ updatetitle(void)
 	printf("\x1b]2;(%lu/%lu) - sfeed_curses\x1b\\", totalnew, total);
 }
 
+void
+appmode(int on)
+{
+	/*fputs(on ? "\x1b[?1049h" : "\x1b[?1049l", stdout);*/ /* smcup, rmcup */
+	putp(tparm(on ? enter_ca_mode : exit_ca_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+}
+
 void
 mousemode(int on)
 {
-	putp(on ? "\x1b[?1000h" : "\x1b[?1000l");
+	fputs(on ? "\x1b[?1000h" : "\x1b[?1000l", stdout); /* xterm mouse mode */
+}
+
+void
+cursormode(int on)
+{
+	/*fputs(on ? "\x1b[?25h" : "\x1b[?25l", stdout);*/ /* DECTCEM (in)Visible cursor */
+	putp(tparm(on ? cursor_visible : cursor_invisible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+}
+
+void
+cursorsave(void)
+{
+	/*fputs("\x1b""7", stdout);*/
+	putp(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+}
+
+void
+cursorrestore(void)
+{
+	/*fputs("\x1b""8", stdout);*/
+	putp(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+}
+
+void
+cursormove(int x, int y)
+{
+	/*printf("\x1b[%d;%dH", y + 1, x + 1);*/
+	putp(tparm(cursor_address, y, x, 0, 0, 0, 0, 0, 0, 0));
+}
+
+void
+attrmode(int mode)
+{
+	char *p;
+
+	/*printf("\x1b[%dm", mode);*/
+	switch (mode) {
+	case ATTR_RESET: p = exit_attribute_mode; break;
+	case ATTR_BOLD_ON: p = enter_bold_mode; break;
+	case ATTR_FAINT_ON: p = enter_dim_mode; break;
+	case ATTR_REVERSE_ON: p = enter_standout_mode; break;
+	case ATTR_REVERSE_OFF: p = exit_standout_mode; break;
+	default: return;
+	}
+	putp(tparm(p, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+}
+
+void
+cleareol(void)
+{
+	/*fputs("\x1b[K", stdout);*/
+	putp(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+}
+
+void
+clearscreen(void)
+{
+	/*fputs("\x1b[H\x1b[2J", stdout);*/
+	putp(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 }
 
 void
 cleanup(void)
 {
+	struct sigaction sa;
+
 	if (!needcleanup)
 		return;
 
 	/* restore terminal settings */
-	tcsetattr(ttyfd, TCSANOW, &tsave);
+	tcsetattr(0, TCSANOW, &tsave);
 
-	putp(tparm(cursor_visible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	putp(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	appmode(0);
+	cursormode(1);
+	clearscreen();
 
 	/* xterm mouse-mode */
 	if (usemouse)
@@ -370,7 +439,10 @@ cleanup(void)
 	resettitle();
 	fflush(stdout);
 
-	signal(SIGWINCH, SIG_DFL);
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART; /* require BSD signal semantics */
+	sa.sa_handler = SIG_DFL;
+	sigaction(SIGWINCH, &sa, NULL);
 
 	needcleanup = 0;
 }
@@ -385,11 +457,16 @@ win_update(struct win *w, int width, int height)
 }
 
 void
-getwinsize(void)
+resizewin(void)
 {
-	if (ioctl(ttyfd, TIOCGWINSZ, &winsz) == -1)
+	/*struct winsize winsz;
+	if (ioctl(0, TIOCGWINSZ, &winsz) == -1)
 		err(1, "ioctl");
-	win_update(&win, winsz.ws_col, winsz.ws_row);
+	win_update(&win, winsz.ws_col, winsz.ws_row);*/
+
+	setupterm(NULL, 1, NULL);
+	/* termios globals are changed: `lines` and `columns` */
+	win_update(&win, columns, lines);
 	if (win.dirty)
 		alldirty();
 }
@@ -397,44 +474,58 @@ getwinsize(void)
 void
 init(void)
 {
-	tcgetattr(ttyfd, &tsave);
+	struct sigaction sa;
+
+	tcgetattr(0, &tsave);
 	memcpy(&tcur, &tsave, sizeof(tcur));
 	tcur.c_lflag &= ~(ECHO|ICANON);
-	tcsetattr(ttyfd, TCSANOW, &tcur);
+	tcsetattr(0, TCSANOW, &tcur);
 
-	getwinsize();
+	resizewin();
 
-	setupterm(NULL, 1, NULL);
-	putp(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	putp(tparm(cursor_invisible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	cursormode(0);
+	appmode(1);
 
 	/* xterm mouse-mode */
 	if (usemouse)
 		mousemode(usemouse);
 
 	updategeom();
-
 	fflush(stdout);
 
-	signal(SIGWINCH, sighandler);
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART; /* require BSD signal semantics */
+	sa.sa_handler = sighandler;
+	sigaction(SIGINT, &sa, NULL);
+	sigaction(SIGTERM, &sa, NULL);
+	sigaction(SIGWINCH, &sa, NULL);
 
 	needcleanup = 1;
 }
 
+/* pipe item to a program, if `wantoutput` is set then cleanup and restore the
+   terminal attribute settings, if not set then don't do that and also ignore
+   stdout and stderr. */
 void
-pipeitem(const char *cmd, struct item *item)
+pipeitem(const char *cmd, struct item *item, int wantoutput)
 {
 	FILE *fp;
 	int i, pid, wpid;
 
-	cleanup();
+	if (wantoutput)
+		cleanup();
 	switch ((pid = fork())) {
 	case -1:
 		err(1, "fork");
 		return;
 	case 0:
+		if (!wantoutput) {
+			dup2(devnullfd, 1);
+			dup2(devnullfd, 2);
+		}
+
 		errno = 0;
-		if (!(fp = popen(cmd, "we"))) {
+		if (!(fp = popen(cmd, "w"))) {
 			fputs("popen: ", stderr);
 			perror(NULL);
 			_exit(1);
@@ -451,9 +542,12 @@ pipeitem(const char *cmd, struct item *item)
 		while ((wpid = wait(NULL)) >= 0 && wpid != pid)
 			;
 	}
-	updatesidebar(onlynew);
-	updatetitle();
-	init();
+
+	if (wantoutput) {
+		updatesidebar(onlynew);
+		updatetitle();
+		init();
+	}
 }
 
 void
@@ -510,16 +604,17 @@ pane_row_draw(struct pane *p, off_t pos)
 	row = pane_row_get(p, pos);
 
 	y = p->y + (pos % p->height); /* relative position on screen */
-	putp(tparm(cursor_address, y, p->x, 0, 0, 0, 0, 0, 0, 0));
+	cursorsave();
+	cursormove(p->x, y);
 
 	r = 0;
 	if (pos == p->pos) {
-		putp(tparm(enter_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		attrmode(ATTR_REVERSE_ON);
 		if (!p->focused)
-			putp(tparm(enter_dim_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+			attrmode(ATTR_FAINT_ON);
 		r = 1;
 	} else if (p->nrows && pos < p->nrows && row && row->bold) {
-		putp(tparm(enter_bold_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		attrmode(ATTR_BOLD_ON);
 		r = 1;
 	}
 	if (row)
@@ -527,7 +622,8 @@ pane_row_draw(struct pane *p, off_t pos)
 	else
 		printf("%-*.*s", p->width, p->width, "");
 	if (r)
-		putp("\x1b[0m");
+		attrmode(ATTR_RESET);
+	cursorrestore();
 }
 
 void
@@ -544,19 +640,19 @@ pane_setpos(struct pane *p, off_t pos)
 	if (pos >= p->nrows)
 		pos = p->nrows - 1; /* clamp */
 
-	/* is on different screen ? */
+	/* is on different scroll region? mark dirty */
 	if (((p->pos - (p->pos % p->height)) / p->height) !=
 	    ((pos - (pos % p->height)) / p->height)) {
 		p->pos = pos;
-		// TODO: dirty region handling? or setting dirty per row?
 		p->dirty = 1;
 		pane_draw(p);
 	} else {
+		/* only redraw the 1 or 2 dirty rows */
 		prev = p->pos;
 		p->pos = pos;
-		// TODO: dirty region handling? or setting dirty per row?
 		pane_row_draw(p, prev); /* draw previous row again */
 		pane_row_draw(p, p->pos); /* draw new highlighted row */
+		fflush(stdout); /* flush and update directly */
 	}
 }
 
@@ -569,7 +665,7 @@ pane_setstartpos(struct pane *p)
 void
 pane_setendpos(struct pane *p)
 {
-	pane_setpos(p, p->nrows);
+	pane_setpos(p, p->nrows - 1);
 }
 
 void
@@ -593,10 +689,7 @@ pane_scrollpage(struct pane *p, int pages)
 void
 pane_scrolln(struct pane *p, int n)
 {
-	if (n < 0)
-		pane_setpos(p, p->pos + n);
-	else if (n > 0)
-		pane_setpos(p, p->pos + n);
+	pane_setpos(p, p->pos + n);
 }
 
 void
@@ -746,69 +839,102 @@ scrollbar_draw(struct scrollbar *s)
 	if (s->hidden || !s->dirty)
 		return;
 
+	cursorsave();
 	if (!s->focused)
-		putp(tparm(enter_dim_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		attrmode(ATTR_FAINT_ON);
 	for (y = 0; y < s->size; y++) {
-		putp(tparm(cursor_address, s->y + y, s->x, 0, 0, 0, 0, 0, 0, 0));
+		cursormove(s->x, s->y + y);
 		if (y >= s->tickpos && y < s->tickpos + s->ticksize) {
-			putp(tparm(enter_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+			attrmode(ATTR_REVERSE_ON);
 			fputs(" ", stdout);
-			putp(tparm(exit_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+			attrmode(ATTR_REVERSE_OFF);
 		} else {
-			fputs("\xe2\x94\x82", stdout); /* symbol: "light vertical" */
+			fputs(SCROLLBAR_SYMBOL, stdout);
 		}
 	}
 	if (!s->focused)
-		putp("\x1b[0m");
+		attrmode(ATTR_RESET);
+	cursorrestore();
 	s->dirty = 0;
 }
 
+int
+readch(void)
+{
+	unsigned char b;
+	ssize_t n;
+
+	n = read(0, &b, 1);
+	if (n == 0)
+		return EOF;
+	else if (n == -1)
+		err(1, "read");
+
+	return (int)b;
+}
+
+char *
+lineeditor(void)
+{
+	char *input = NULL;
+	size_t cap = 0, nchars = 0;
+	int ch;
+
+	for (;;) {
+		if (nchars + 1 >= cap) {
+			cap = cap ? cap * 2 : 32;
+			input = erealloc(input, cap);
+		}
+
+		ch = readch();
+		if (ch == EOF || ch == '\r' || ch == '\n') {
+			input[nchars] = '\0';
+			break;
+		} else if (ch == '\b' || ch == 0x7f) {
+			if (!nchars)
+				continue;
+			input[--nchars] = '\0';
+			ch = '\b'; /* back */
+			write(1, &ch, 1);
+			ch = ' '; /* blank */
+			write(1, &ch, 1);
+			ch = '\b'; /* back */
+			write(1, &ch, 1);
+			continue;
+		} else if (ch >= ' ') {
+			write(1, &ch, 1);
+		}
+		input[nchars++] = ch;
+	}
+	return input;
+}
+
 char *
 uiprompt(int x, int y, char *fmt, ...)
 {
 	va_list ap;
-	char *input = NULL;
-	size_t n;
-	ssize_t r;
+	char *input;
 	char buf[32];
-	struct termios tset;
 
 	va_start(ap, fmt);
 	if (vsnprintf(buf, sizeof(buf), fmt, ap) >= sizeof(buf))
 		buf[sizeof(buf) - 1] = '\0';
 	va_end(ap);
 
-	tcgetattr(ttyfd, &tset);
-	tset.c_lflag |= (ECHO|ICANON);
-	tcsetattr(ttyfd, TCSANOW, &tset);
-
-	putp(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	putp(tparm(cursor_address, y, x, 0, 0, 0, 0, 0, 0, 0));
-	putp(tparm(enter_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	cursorsave();
+	cursormove(x, y);
+	attrmode(ATTR_REVERSE_ON);
 	fputs(buf, stdout);
-	putp(tparm(exit_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-
-	putp(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	putp(tparm(cursor_visible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	putp(tparm(cursor_address, y, x + colw(buf), 0, 0, 0, 0, 0, 0, 0));
+	attrmode(ATTR_REVERSE_OFF);
+	cleareol();
+	cursormode(1);
+	cursormove(x + colw(buf) + 1, y);
 	fflush(stdout);
 
-	n = 0;
-	r = getline(&input, &n, stdin);
+	input = lineeditor();
 
-	putp(tparm(cursor_invisible, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	tcsetattr(ttyfd, TCSANOW, &tcur); /* restore */
-	putp(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-
-	fflush(stdout);
-
-	if (r < 0) {
-		clearerr(stdin);
-		free(input);
-		input = NULL;
-	} else if (input[r - 1] == '\n') {
-		input[--r] = '\0';
-	}
+	cursormode(0);
+	cursorrestore();
 
 	return input;
 }
@@ -818,10 +944,13 @@ statusbar_draw(struct statusbar *s)
 {
 	if (s->hidden || !s->dirty)
 		return;
-	putp(tparm(cursor_address, s->y, s->x, 0, 0, 0, 0, 0, 0, 0));
-	putp(tparm(enter_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+
+	cursorsave();
+	cursormove(s->x, s->y);
+	attrmode(ATTR_REVERSE_ON);
 	printpad(s->text, s->width);
-	putp(tparm(exit_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+	attrmode(ATTR_REVERSE_OFF);
+	cursorrestore();
 	s->dirty = 0;
 }
 
@@ -841,21 +970,21 @@ int
 linetoitem(char *line, struct item *item)
 {
 	char *fields[FieldLast];
-	struct tm tm;
 	time_t parsedtime;
 
 	parseline(line, fields);
 
 	parsedtime = 0;
-	if (strtotime(fields[FieldUnixTimestamp], &parsedtime))
-		return -1;
-	if (!localtime_r(&parsedtime, &tm))
-		return -1;
+	if (!strtotime(fields[FieldUnixTimestamp], &parsedtime)) {
+		item->timestamp = parsedtime;
+		item->timeok = 1;
+	} else {
+		item->timestamp = 0;
+		item->timeok = 0;
+	}
 
 	item->line = line;
 	item->isnew = (parsedtime >= comparetime);
-	item->timestamp = parsedtime;
-	memcpy(&(item->tm), &tm, sizeof(tm));
 	memcpy(item->fields, fields, sizeof(fields));
 
 	return 0;
@@ -865,7 +994,7 @@ int
 feed_getitems(struct feed *f, FILE *fp, struct item **items, size_t *nitems)
 {
 	struct item *item;
-	char *dupline, *line = NULL;
+	char *line = NULL;
 	size_t cap, i, linesize = 0;
 	ssize_t linelen;
 	off_t offset;
@@ -878,10 +1007,7 @@ feed_getitems(struct feed *f, FILE *fp, struct item **items, size_t *nitems)
 	offset = 0;
 	for (i = 0; ; i++) {
 		if (i + 1 >= cap) {
-			if (cap == 0)
-				cap = 16;
-			else
-				cap *= 2;
+			cap = cap ? cap * 2 : 16;
 			*items = erealloc(*items, cap * sizeof(struct item));
 		}
 		if ((linelen = getline(&line, &linesize, fp)) > 0) {
@@ -894,20 +1020,13 @@ feed_getitems(struct feed *f, FILE *fp, struct item **items, size_t *nitems)
 				line[--linelen] = '\0';
 
 			if (lazyload && f->path) {
-				if (linetoitem(line, item) == -1) {
-					i--;
-					continue;
-				}
+				linetoitem(line, item);
+
 				/* data is ignored here, will be lazy-loaded later. */
 				item->line = NULL;
 				memset(item->fields, 0, sizeof(item->fields));
 			} else {
-				dupline = estrdup(line);
-				if (linetoitem(dupline, item) == -1) {
-					i--;
-					free(dupline);
-					continue;
-				}
+				linetoitem(estrdup(line), item);
 			}
 
 			(*nitems)++;
@@ -972,7 +1091,6 @@ feed_count(struct feed *f, FILE *fp)
 	char *line = NULL;
 	size_t linesize = 0;
 	ssize_t linelen;
-	struct tm *tm;
 	time_t parsedtime;
 
 	while ((linelen = getline(&line, &linesize, fp)) > 0) {
@@ -981,12 +1099,8 @@ feed_count(struct feed *f, FILE *fp)
 		parseline(line, fields);
 
 		parsedtime = 0;
-		if (strtotime(fields[FieldUnixTimestamp], &parsedtime))
-			continue;
-		if (!(tm = localtime(&parsedtime)))
-			continue; /* ignore item */
-
-		f->totalnew += (parsedtime >= comparetime);
+		if (!strtotime(fields[FieldUnixTimestamp], &parsedtime))
+			f->totalnew += (parsedtime >= comparetime);
 		f->total++;
 	}
 	free(line);
@@ -1053,10 +1167,6 @@ feeds_load(struct feed *feeds, size_t nfeeds)
 		else
 			feed_count(f, f->fp);
 
-		/* must be able to seek if reading from a path */
-		if (lazyload && f->path && fseek(f->fp, 0, SEEK_SET))
-			err(1, "fseek: %s", f->path);
-
 		if (f->path && f->fp) {
 			fclose(f->fp);
 			f->fp = NULL;
@@ -1099,15 +1209,23 @@ updatesidebar(int onlynew)
 		p->nrows++;
 	}
 	p->width = width;
+	p->dirty = 1;
 }
 
 void
 sighandler(int signo)
 {
-	if (signo == SIGWINCH) {
-		getwinsize();
+	switch (signo) {
+	case SIGINT:
+	case SIGTERM:
+		cleanup();
+		_exit(128 + signo);
+		break;
+	case SIGWINCH:
+		resizewin();
 		updategeom();
 		draw();
+		break;
 	}
 }
 
@@ -1134,7 +1252,7 @@ draw(void)
 		return;
 
 	if (win.dirty) {
-		putp(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+		clearscreen();
 		win.dirty = 0;
 	}
 
@@ -1170,7 +1288,7 @@ mousereport(int button, int release, int x, int y)
 	struct row *row;
 	struct item *item;
 	size_t i;
-	int changedpane, pos;
+	int changedpane, dblclick, pos;
 
 	if (!usemouse || release || button == -1)
 		return;
@@ -1188,13 +1306,14 @@ mousereport(int button, int release, int x, int y)
 		selpane = i;
 		/* relative position on screen */
 		pos = y - p->y + p->pos - (p->pos % p->height);
+		dblclick = (pos == p->pos); /* clicking the same row */
 
 		switch (button) {
 		case 0: /* left-click */
 			if (!p->nrows || pos >= p->nrows)
 				break;
+			pane_setpos(p, pos);
 			if (i == PaneFeeds) {
-				pane_setpos(p, pos);
 				row = pane_row_get(p, pos);
 				f = (struct feed *)row->data;
 				feeds_set(f);
@@ -1203,25 +1322,22 @@ mousereport(int button, int release, int x, int y)
 				pane_row_draw(p, pos);
 				updatetitle();
 			} else if (i == PaneItems) {
-				/* clicking the same highlighted row */
-				if (p->pos == pos && !changedpane) {
+				if (dblclick && !changedpane) {
 					row = pane_row_get(&panes[PaneItems], pos);
 					item = (struct item *)row->data;
 					plumb(plumber, item->fields[FieldLink]);
-				} else {
-					pane_setpos(p, pos);
 				}
 			}
 			break;
 		case 2: /* right-click */
 			if (!p->nrows || pos >= p->nrows)
 				break;
+			pane_setpos(p, pos);
 			if (i == PaneItems) {
-				pane_setpos(p, pos);
 				p = &panes[PaneItems];
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
-				pipeitem(piper, item);
+				pipeitem(piper, item, 1);
 			}
 			break;
 		case 3: /* scroll up */
@@ -1229,7 +1345,6 @@ mousereport(int button, int release, int x, int y)
 			pane_scrollpage(p, button == 3 ? -1 : +1);
 			break;
 		}
-		draw();
 	}
 }
 
@@ -1273,7 +1388,7 @@ item_row_get(struct pane *p, off_t pos)
 	struct row *itemrow;
 	struct item *item;
 	struct feed *f;
-	char *dupline, *line = NULL;
+	char *line = NULL;
 	size_t linesize = 0;
 	ssize_t linelen;
 
@@ -1292,11 +1407,7 @@ item_row_get(struct pane *p, off_t pos)
 		if (line[linelen - 1] == '\n')
 			line[--linelen] = '\0';
 
-		dupline = estrdup(line);
-		if (linetoitem(dupline, item) == -1) {
-			free(dupline);
-			return NULL;
-		}
+		linetoitem(estrdup(line), item);
 		free(line);
 
 		itemrow->data = item;
@@ -1308,15 +1419,23 @@ item_row_get(struct pane *p, off_t pos)
 char *
 item_row_format(struct pane *p, struct row *row)
 {
-	struct item *item;
 	static char text[1024];
+	struct item *item;
+	struct tm tm;
 
 	item = (struct item *)row->data;
 
-	snprintf(text, sizeof(text), "%c %04d-%02d-%02d %02d:%02d %s",
-	         item->fields[FieldEnclosure][0] ? '@' : ' ',
-	         item->tm.tm_year + 1900, item->tm.tm_mon + 1, item->tm.tm_mday,
-	         item->tm.tm_hour, item->tm.tm_min, item->fields[FieldTitle]);
+	if (item->timeok && localtime_r(&(item->timestamp), &tm)) {
+		snprintf(text, sizeof(text), "%c %04d-%02d-%02d %02d:%02d %s",
+		         item->fields[FieldEnclosure][0] ? '@' : ' ',
+		         tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
+		         tm.tm_hour, tm.tm_min, item->fields[FieldTitle]);
+	} else {
+		localtime_r(&(item->timestamp), &tm);
+		snprintf(text, sizeof(text), "%c                  %s",
+		         item->fields[FieldEnclosure][0] ? '@' : ' ',
+		         item->fields[FieldTitle]);
+	}
 
 	return text;
 }
@@ -1331,7 +1450,7 @@ main(int argc, char *argv[])
 	size_t i;
 	char *name, *tmp;
 	char *search = NULL; /* search text */
-	int ch, button, x, y, release; /* mouse button event */
+	int ch, button, fd, x, y, release;
 	off_t off;
 
 	setlocale(LC_CTYPE, "");
@@ -1346,7 +1465,7 @@ main(int argc, char *argv[])
 		nfeeds = 1;
 		f = &feeds[0];
 		f->name = "stdin";
-		if (!(f->fp = fdopen(ttyfd, "rb")))
+		if (!(f->fp = fdopen(0, "rb")))
 			err(1, "fdopen");
 	} else {
 		for (i = 1; i < argc; i++) {
@@ -1361,11 +1480,10 @@ main(int argc, char *argv[])
 	feeds_load(feeds, nfeeds);
 	feeds_set(&feeds[0]);
 
-	if (!isatty(ttyfd)) {
-		close(ttyfd);
-		if ((ttyfd = open("/dev/tty", O_RDONLY)) == -1)
+	if (!isatty(0)) {
+		if ((fd = open("/dev/tty", O_RDONLY)) == -1)
 			err(1, "open: /dev/tty");
-		if (dup2(ttyfd, 0) == -1)
+		if (dup2(fd, 0) == -1)
 			err(1, "dup2: /dev/tty");
 	}
 	if (argc == 1)
@@ -1390,40 +1508,37 @@ main(int argc, char *argv[])
 	updatetitle();
 	updatesidebar(onlynew);
 	init();
-
-	atexit(cleanup);
-
 	draw();
 
-	while ((ch = getchar()) != EOF) {
+	while ((ch = readch()) != EOF) {
 		switch (ch) {
 		case '\x1b':
-			if ((ch = getchar()) != '[')
+			ch = readch();
+			if (ch != '[' && ch != 'O')
 				continue; /* unhandled */
-			switch ((ch = getchar())) {
+			switch ((ch = readch())) {
 			case EOF: goto end;
 			case 'M': /* reported mouse event */
-				if ((ch = getchar()) == EOF)
+				if ((ch = readch()) == EOF)
 					goto end;
 				/* button numbers (0 - 2) encoded in lowest 2 bits
-				   release does not indicate which button (so set to 0). */
-				ch -= 32;
+				   release does not indicate which button (so set to 0).
+				   Handle extended buttons like scrollwheels
+				   and side-buttons by substracting 64 in each range. */
+				for (i = 0, ch -= 32; ch >= 64; i += 3)
+					ch -= 64;
 
-				/* extended buttons like scrollwheels */
 				release = 0;
-				if (ch >= 64) {
-					button = ((ch - 64) & 3) + 3;
-				} else {
-					button = ch & 3;
-					if (button == 3) {
-						button = -1;
-						release = 1;
-					}
+				button = (ch & 3) + i;
+				if (!i && button == 3) {
+					release = 1;
+					button = -1;
 				}
+
 				/* X10 mouse-encoding */
-				if ((x = getchar()) == EOF)
+				if ((x = readch()) == EOF)
 					goto end;
-				if ((y = getchar()) == EOF)
+				if ((y = readch()) == EOF)
 					goto end;
 				mousereport(button, release, x - 33, y - 33);
 				break;
@@ -1431,21 +1546,22 @@ main(int argc, char *argv[])
 			case 'B': goto keydown;  /* arrow down */
 			case 'C': goto keyright; /* arrow left */
 			case 'D': goto keyleft;  /* arrow right */
+			case 'F': goto endpos;   /* end */
 			case 'H': goto startpos; /* home */
 			case '4': /* end */
-				if ((ch = getchar()) == '~')
+				if ((ch = readch()) == '~')
 					goto endpos;
 				continue;
 			case '5': /* page up */
-				if ((ch = getchar()) == '~')
+				if ((ch = readch()) == '~')
 					goto prevpage;
 				continue;
 			case '6': /* page down */
-				if ((ch = getchar()) == '~')
+				if ((ch = readch()) == '~')
 					goto nextpage;
 				continue;
 			}
-			continue;
+			break;
 keyup:
 		case 'k':
 			pane_scrolln(&panes[selpane], -1);
@@ -1495,9 +1611,8 @@ nextpage:
 				tmp = ch == '?' ? "backward" : "forward";
 				free(search);
 				search = uiprompt(statusbar.x, statusbar.y,
-				                  "Search (%s): ", tmp);
-				alldirty();
-				draw();
+				                  "Search (%s):", tmp);
+				statusbar.dirty = 1;
 			}
 			if (!search)
 				break;
@@ -1526,7 +1641,7 @@ nextpage:
 			break;
 		case 'R': /* reload all files */
 			if (nfeeds == 1 && !feeds[0].path)
-				break;
+				break; /* do not reload when read from stdin */
 			feeds_set(NULL);
 			feeds_load(feeds, nfeeds);
 			feeds_set(&feeds[0]);
@@ -1560,7 +1675,6 @@ nextpage:
 			onlynew = !onlynew;
 			pane_setpos(&panes[PaneFeeds], 0);
 			updatesidebar(onlynew);
-			updategeom();
 			break;
 		case 'o': /* feeds: load, items: plumb url */
 		case '\n':
@@ -1590,9 +1704,9 @@ nextpage:
 				row = pane_row_get(p, p->pos);
 				item = (struct item *)row->data;
 				switch (ch) {
-				case 'y': pipeitem("cut -f 3 | xclip -r", item); break;
-				case 'E': pipeitem("cut -f 8 | xclip -r", item); break;
-				default:  pipeitem(piper, item); break;
+				case 'y': pipeitem("cut -f 3 | xclip -r", item, 0); break;
+				case 'E': pipeitem("cut -f 8 | xclip -r", item, 0); break;
+				default:  pipeitem(piper, item, 1); break;
 				}
 			}
 			break;
@@ -1602,5 +1716,7 @@ nextpage:
 		draw();
 	}
 end:
+	cleanup();
+
 	return 0;
 }
